/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <config.h>
#include <gnome.h>

#include <sys/types.h>
#include <sys/wait.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>

#include "acache.h"
#include "articlelist.h"
#include "encode.h"
#include "globals.h"
#include "gnksa.h"
#include "grouplist.h"
#include "gui.h"
#include "gui-headers.h"
#include "message-check.h"
#include "message-send.h"
#include "message-window.h"
#include "prefs.h"
#include "print.h"
#include "text.h" /* for colors of fg, bg, and quoted text */
#include "util.h"


extern GtkTooltips *ttips;

static void wrap_cb                        (GtkWidget*, MessageWindow*);
static void save_cb                        (GtkWidget*, MessageWindow*);
static void send_now_cb                    (GtkWidget*, MessageWindow*);
static void send_later_cb                  (GtkWidget*, MessageWindow*);
static void window_destroy_cb              (GtkWidget*, MessageWindow*);
static void message_window_save_cb         (GtkWidget*, MessageWindow*);
static void message_window_cut_cb          (GtkWidget*, MessageWindow*);
static void message_window_copy_cb         (GtkWidget*, MessageWindow*);
static void message_window_paste_cb        (GtkWidget*, MessageWindow*);
static void message_window_close           (GtkWidget*, MessageWindow*);
static void message_window_print           (GtkWidget*, MessageWindow*);
static void message_window_appbar_toggle   (GtkWidget*, MessageWindow*);
static void attachment_file_selector       (GtkWidget*, MessageWindow*);
static GtkWidget* create_read_body_pane    (GtkWidget*, MessageWindowType, Article*);
static GtkWidget* attachment_page          (MessageWindow*);
static GtkWidget* create_post_body_pane    (MessageWindow*);
static void attachments_lines_per_part_changed (GtkWidget*, MessageWindow*);

/*static void message_edit_external (GtkWidget *widget, Message *msg);*/


gboolean use_message_window_always = FALSE;

/***
****
***/

static GnomeUIInfo message_edit_file_menu[] =
{
	GNOMEUIINFO_ITEM_STOCK (N_("Save Changes"), N_("Save Changes"), 
				save_cb, GNOME_STOCK_MENU_SAVE),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_MENU_SAVE_AS_ITEM (message_window_save_cb, NULL),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_MENU_CLOSE_ITEM (message_window_close, NULL),	
	GNOMEUIINFO_END
};

static GnomeUIInfo message_post_file_menu[] =
{
	GNOMEUIINFO_ITEM_STOCK (N_("Send Now"), N_("Send this message now."), 
				send_now_cb, GNOME_STOCK_MENU_MAIL_SND),
	GNOMEUIINFO_ITEM_STOCK (N_("Send Later"), N_("Send this message now."), 
				send_later_cb, GNOME_STOCK_MENU_TIMER),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_MENU_SAVE_AS_ITEM (message_window_save_cb, NULL),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_MENU_CLOSE_ITEM (message_window_close, NULL),	
	GNOMEUIINFO_END
};

static GnomeUIInfo message_read_file_menu[] =
{
	GNOMEUIINFO_MENU_SAVE_AS_ITEM (message_window_save_cb, NULL),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_MENU_PRINT_ITEM (message_window_print, NULL),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_MENU_CLOSE_ITEM (message_window_close, NULL),	
	GNOMEUIINFO_END
};

static GnomeUIInfo message_edit_menu[] =
{
	GNOMEUIINFO_MENU_CUT_ITEM (message_window_cut_cb, NULL),
	GNOMEUIINFO_MENU_COPY_ITEM (message_window_copy_cb, NULL),
	GNOMEUIINFO_MENU_PASTE_ITEM (message_window_paste_cb, NULL),
	GNOMEUIINFO_END
};

static GnomeUIInfo message_view_menu[] =
{
	{
		GNOME_APP_UI_TOGGLEITEM,
		N_ ("Show Status Bar"), 0,
		message_window_appbar_toggle, 0, 0, 0, 0, 0, 0, 0
	},
	GNOMEUIINFO_END
};

static GnomeUIInfo message_read_message_menu[] =
{
	{
		GNOME_APP_UI_ITEM,
		N_("_Post to newsgroup"),
		N_("Post a new message to the current group."),
		message_post_window
	},
	{
		GNOME_APP_UI_ITEM,
		N_("_Followup to newsgroup"),
		N_("Post a reply to the message on the news server."),
		message_followup_window
	},
	{
		GNOME_APP_UI_ITEM,
		N_("Reply by _E-mail"),
		N_("Create a mail reply to the sender."),
		message_reply_window
	},
	{
		GNOME_APP_UI_ITEM,
		N_("Reply with External Mailer"),
		N_("Create a mail reply to the sender with your favorite MUA."),
		message_reply_external,
		NULL
       	},
	GNOMEUIINFO_END
};

static GnomeUIInfo message_post_message_menu[] =
{
	{ GNOME_APP_UI_ITEM, N_("New"), N_("Create a new message."), message_post_window },
	GNOMEUIINFO_END
};

static GnomeUIInfo message_read_main_menu[] =
{
	GNOMEUIINFO_MENU_FILE_TREE (message_read_file_menu),
	GNOMEUIINFO_MENU_EDIT_TREE (message_edit_menu),
	GNOMEUIINFO_SUBTREE (N_("_View"), &message_view_menu),
	GNOMEUIINFO_SUBTREE (N_("_Message"), &message_read_message_menu),
	GNOMEUIINFO_MENU_HELP_TREE (help_menu),
	GNOMEUIINFO_END
};

static GnomeUIInfo message_post_main_menu[] =
{
	GNOMEUIINFO_MENU_FILE_TREE (message_post_file_menu),
	GNOMEUIINFO_MENU_EDIT_TREE (message_edit_menu),
	GNOMEUIINFO_SUBTREE (N_("_View"), &message_view_menu),
	GNOMEUIINFO_SUBTREE (N_("_Message"), &message_post_message_menu),
	GNOMEUIINFO_MENU_HELP_TREE (help_menu),
	GNOMEUIINFO_END
};

static GnomeUIInfo message_edit_main_menu[] =
{
	GNOMEUIINFO_MENU_FILE_TREE (message_edit_file_menu),
	GNOMEUIINFO_MENU_EDIT_TREE (message_edit_menu),
	GNOMEUIINFO_SUBTREE (N_("_View"), &message_view_menu),
	GNOMEUIINFO_MENU_HELP_TREE (help_menu),
	GNOMEUIINFO_END
};


static GnomeUIInfo message_read_toolbar[] =
{
	GNOMEUIINFO_ITEM_STOCK(
		N_("Post Followup"),
		N_("Post a reply to the message on the news server."),
		message_followup_window,
		GNOME_STOCK_PIXMAP_MAIL_NEW),
	GNOMEUIINFO_ITEM_STOCK(
		N_("E-mail Reply"),
		N_("Create a mail reply to the sender."),
		message_reply_window,
		GNOME_STOCK_PIXMAP_MAIL_RPL),
#if 0
	GNOMEUIINFO_ITEM_STOCK (N_("Forward"), N_("Forward this Message"), 
				NULL, GNOME_STOCK_PIXMAP_MAIL_FWD),
#endif
	GNOMEUIINFO_SEPARATOR,
#if 0
	GNOMEUIINFO_ITEM_STOCK (N_("Spelling"), N_("Check spelling"),
				NULL, GNOME_STOCK_PIXMAP_SPELLCHECK),
#endif
	GNOMEUIINFO_ITEM_STOCK (N_("Print"), N_("Print this message"),
				message_window_print, GNOME_STOCK_PIXMAP_PRINT),

	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_ITEM_STOCK (N_("Close"), N_("Close this Window"),
				message_window_close, GNOME_STOCK_PIXMAP_CLOSE),
	GNOMEUIINFO_END
};

static GnomeUIInfo message_edit_toolbar[] =
{
	GNOMEUIINFO_ITEM_STOCK
	(
		N_("Save Changes"),
		N_("Save Changes"),
		save_cb,
		GNOME_STOCK_PIXMAP_SAVE
	),
	GNOMEUIINFO_END
};

static GnomeUIInfo message_post_toolbar[] =
{
	GNOMEUIINFO_ITEM_STOCK
	(
		N_("Send Now"),
		N_("Send this Message Now"), 
		send_now_cb,
		GNOME_STOCK_PIXMAP_MAIL_SND
	),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_ITEM_STOCK
	(
		N_("Send Later"),
		N_("Send this Message Later"), 
		send_later_cb,
		GNOME_STOCK_PIXMAP_TIMER
	),
	GNOMEUIINFO_SEPARATOR,
#if 0
	GNOMEUIINFO_ITEM_STOCK
	(
		N_("Spelling"),
		N_("Check spelling"),
		NULL,
		GNOME_STOCK_PIXMAP_SPELLCHECK
	),
	GNOMEUIINFO_ITEM_STOCK
	(
		N_("Print"),
		N_("Print this message"),
		not_implemented,
		GNOME_STOCK_PIXMAP_PRINT
	),
	GNOMEUIINFO_SEPARATOR,
#endif
	GNOMEUIINFO_ITEM_STOCK
	(
		N_("Wrap Text"),
		N_("Wrap the message body"),
		wrap_cb,
		GNOME_STOCK_PIXMAP_ALIGN_JUSTIFY
	),
#if 0
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_ITEM_STOCK
	(
		N_("Open Editor"),
		N_("Edit this in your favorite editor"),
		message_edit_external,
		GNOME_STOCK_PIXMAP_DOWN
	),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_ITEM_STOCK
	(
		N_("Close"),
		N_("Close this Window without Sending Message"),
		message_window_close,
		GNOME_STOCK_PIXMAP_CLOSE
	),
#endif
	GNOMEUIINFO_END
};

/***
****
***/

static gboolean
is_posting (MessageWindowType type)
{
	return type==NNTP_POST
		|| type==NNTP_REPLY
		|| type==EMAIL_AND_POST_REPLY;
}

static gchar*
make_reply_string (const gchar* string)
{
	gchar * retval;

	if (g_strncasecmp ("Re: ", string, 4))
		retval = g_strconcat ("Re: ", string, NULL);
	else
		retval = g_strdup (string);

	return retval;
}

static void
message_window_appbar_toggle (GtkWidget *widget, MessageWindow * mw)
{
	pan_lock ();

	if (mw->appbar == NULL)
	{
		mw->appbar = gnome_appbar_new (0, 1, GNOME_PREFERENCES_USER);
		gnome_app_set_statusbar (GNOME_APP(mw->window),
		                         GTK_WIDGET(mw->appbar));
		gnome_app_install_menu_hints (GNOME_APP (mw->window),
			(mw->type==NNTP_READ
			 ? message_read_main_menu
			 : message_post_main_menu));
	}
	else if (GTK_IS_WIDGET (mw->appbar->parent))
	{
		gtk_widget_ref (mw->appbar);
		gtk_container_remove (GTK_CONTAINER(mw->appbar->parent),
		                      mw->appbar);
	}
	else
	{
		gtk_box_pack_start (GTK_BOX(GNOME_APP(mw->window)->vbox),
		                    mw->appbar, FALSE, FALSE, 0);
		gtk_widget_unref (mw->appbar);
	}
	pan_unlock ();
}



#if 0
static void
message_edit_external (GtkWidget *widget, Message *msg)
{
	not_implemented ();
}
#endif

static void
wrap_cb (GtkWidget *widget, MessageWindow * mw)
{
	gchar * body;
	gchar * new_body;

	/* get the current body */
	pan_lock();
       	body = gtk_editable_get_chars (GTK_EDITABLE(mw->body), 0, -1);
	pan_unlock ();

	/* fill the text, and update the body */
	new_body = fill_body (body);
	g_free (body);
	update_body_pane (mw->body, new_body);
	g_free (new_body);
}

/**
***
**/

static void
message_window_cut_cb (GtkWidget * widget, MessageWindow * mw)
{
	g_return_if_fail (mw->body);

	pan_lock ();
	gtk_editable_cut_clipboard (GTK_EDITABLE(mw->body));
	pan_unlock ();
}


static void
message_window_copy_cb (GtkWidget * widget, MessageWindow * mw)
{
	g_return_if_fail (mw->body);

	pan_lock ();
	gtk_editable_copy_clipboard (GTK_EDITABLE(mw->body));
	pan_unlock ();
}


static void
message_window_paste_cb (GtkWidget *widget, MessageWindow * mw)
{
	pan_lock();
	gtk_editable_paste_clipboard (GTK_EDITABLE(mw->body));
	pan_unlock();
}


static void
message_window_destroy (MessageWindow *mw)
{
	pan_lock();
/*	gtk_container_foreach (GTK_CONTAINER (mw->window), GTK_SIGNAL_FUNC (gtk_widget_destroy), NULL); */
	gtk_widget_destroy (mw->window);
	pan_unlock();
}

static void
message_window_print (GtkWidget * widget, MessageWindow * mw)
{
	print_cb (widget, mw->article);
}

static void
message_window_close (GtkWidget * widget, MessageWindow * mw)
{
	message_window_destroy (mw);
}


static GtkWidget *
create_read_info_pane (MessageWindow *mw)
{
	GtkWidget * table;

	/* create the table */
	pan_lock();
	table = gtk_table_new (7, 2, FALSE);
	gtk_table_set_row_spacings (GTK_TABLE(table), GNOME_PAD_SMALL);
	gtk_table_set_col_spacings (GTK_TABLE(table), GNOME_PAD_SMALL);
	pan_unlock();

	/* populate the table */
	gui_headers_set_default (table, mw->article);

	return table;
}


static void
update_title_with_subject (GtkWidget * widget, MessageWindow * mw)
{
	gtk_window_set_title (GTK_WINDOW (mw->window),
	                      gtk_entry_get_text (GTK_ENTRY (widget)));
}

static GtkWidget *
attachment_page (MessageWindow *mw)
{
	GtkWidget *table;
	GtkWidget *w;
	GtkObject *a;

	table = gtk_table_new (6, 2, FALSE);
	gtk_table_set_row_spacings (GTK_TABLE(table), GNOME_PAD_SMALL);
	gtk_table_set_col_spacings (GTK_TABLE(table), GNOME_PAD_SMALL);

	w = gtk_label_new (_("Filename:"));
	gtk_misc_set_alignment (GTK_MISC (w), 0.0, 0.5);
	gtk_table_attach (GTK_TABLE (table), w, 0, 1, 0, 1,
                    GTK_FILL, GTK_FILL, GNOME_PAD, 0);

	mw->filename = gtk_entry_new ();
	gtk_table_attach (GTK_TABLE (table), mw->filename, 1, 5, 0, 1,
                    GTK_EXPAND | GTK_FILL, GTK_FILL, GNOME_PAD, 0);

	w = gtk_button_new_with_label (_("Select File"));
	gtk_signal_connect (GTK_OBJECT (w), "clicked",
                      GTK_SIGNAL_FUNC (attachment_file_selector), mw);
	gtk_table_attach (GTK_TABLE (table), w, 5, 6, 0, 1,
                    GTK_FILL, GTK_FILL, GNOME_PAD, 0);

	w = gtk_label_new (_("Lines per Part:"));
	gtk_misc_set_alignment (GTK_MISC (w), 0.0, 0.5);
	gtk_table_attach (GTK_TABLE (table), w, 0, 1, 1, 2,
                    GTK_FILL, GTK_FILL, GNOME_PAD, 0);

	a = gtk_adjustment_new (7000, 100, 15000, 25, 10, 10);
	mw->lines_per_part = gtk_spin_button_new (GTK_ADJUSTMENT (a), 1, 0);
	gtk_table_attach (GTK_TABLE (table), mw->lines_per_part, 1, 2, 1, 2,
                    GTK_EXPAND | GTK_FILL, GTK_FILL, GNOME_PAD, 0);
 	gtk_signal_connect (GTK_OBJECT (a), "value_changed",
                           GTK_SIGNAL_FUNC (attachments_lines_per_part_changed),
                           (gpointer) mw);

	w = gtk_label_new (_("Lines:"));
	gtk_misc_set_alignment (GTK_MISC (w), 0.0, 0.5);
	gtk_table_attach (GTK_TABLE (table), w, 2, 3, 1, 2,
                    GTK_FILL, GTK_FILL, GNOME_PAD, 0);

	mw->lines = gtk_label_new (_("0"));
	gtk_misc_set_alignment (GTK_MISC (mw->lines), 0.0, 0.5);
	gtk_table_attach (GTK_TABLE (table), mw->lines, 3, 4, 1, 2,
                    GTK_EXPAND | GTK_FILL, GTK_FILL, GNOME_PAD, 0);

	w = gtk_label_new (_("Parts:"));
	gtk_misc_set_alignment (GTK_MISC (w), 0.0, 0.5);
	gtk_table_attach (GTK_TABLE (table), w, 4, 5, 1, 2,
                    GTK_FILL, GTK_FILL, GNOME_PAD, 0);

	mw->parts = gtk_label_new (_("0"));
	gtk_misc_set_alignment (GTK_MISC (mw->parts), 0.0, 0.5);
	gtk_table_attach (GTK_TABLE (table), mw->parts, 5, 6, 1, 2,
                    GTK_EXPAND | GTK_FILL, GTK_FILL, GNOME_PAD, 0);

	return table;
}

/*--------------------------------------------------------------------
 * Newsgroups:
 * CC:
 * Subject:
 *--------------------------------------------------------------------*/
static GtkWidget*
create_post_info_pane (MessageWindow *mw)
{
	GtkWidget * headers_page;
	GtkWidget * adv_headers_page;
	GtkWidget * label;
	GtkWidget * eventbox;
	GtkWidget * notebook;

	pan_lock();
	
	/*-----
	 * Headers Page
	 * ---- */
	
	headers_page = gtk_table_new (4, 2, FALSE);
	gtk_table_set_row_spacings (GTK_TABLE(headers_page), GNOME_PAD_SMALL);
	gtk_table_set_col_spacings (GTK_TABLE(headers_page), GNOME_PAD_SMALL);

	label = gtk_label_new ( _("From:") );
	gtk_misc_set_alignment (GTK_MISC(label), 0.0, 0.5);
	gtk_table_attach (GTK_TABLE(headers_page), label, 0, 1, 0, 1,
			  GTK_FILL, GTK_FILL, GNOME_PAD, 0);

	mw->from = gtk_entry_new ();
	gtk_table_attach (GTK_TABLE(headers_page), mw->from, 1, 2, 0, 1,
			  GTK_FILL | GTK_EXPAND, GTK_FILL, GNOME_PAD, 0);

	label = gtk_label_new ( _("Post To Groups:") );
	gtk_misc_set_alignment (GTK_MISC(label), 0.0, 0.5);
	eventbox = gtk_event_box_new ();
	gtk_container_add (GTK_CONTAINER(eventbox), label);
	gtk_tooltips_set_tip (GTK_TOOLTIPS(ttips), eventbox,
						  _("Groups to Post this message to"), "");
	gtk_table_attach (GTK_TABLE(headers_page), eventbox,
	                  0, 1, 1, 2,
			  GTK_FILL, GTK_FILL, GNOME_PAD, 0);
	mw->newsgroups = gtk_entry_new ();
	gtk_table_attach (GTK_TABLE(headers_page), mw->newsgroups,
	                  1, 2, 1, 2,
			  GTK_FILL|GTK_EXPAND, GTK_FILL, GNOME_PAD, 0);



	label = gtk_label_new ( _("Mail To:") );
	gtk_misc_set_alignment (GTK_MISC(label), 0.0, 0.5);
	eventbox = gtk_event_box_new ();
	gtk_container_add (GTK_CONTAINER(eventbox), label);
	gtk_tooltips_set_tip (GTK_TOOLTIPS(ttips), eventbox,
						  _("Users to Mail this message to"), "");
	gtk_table_attach (GTK_TABLE(headers_page), eventbox, 0, 1, 2, 3,
			  GTK_FILL, GTK_FILL, GNOME_PAD, 0);
	mw->users = gtk_entry_new();
	gtk_table_attach (GTK_TABLE(headers_page), mw->users,
	                  1, 2, 2, 3,
			  GTK_FILL | GTK_EXPAND, GTK_FILL, GNOME_PAD, 0);





	label = gtk_label_new ( _("Subject:") );
	gtk_misc_set_alignment (GTK_MISC(label), 0.0, 0.5);
	gtk_table_attach (GTK_TABLE(headers_page), label, 0, 1, 3, 4,
			  GTK_FILL, GTK_FILL, GNOME_PAD, 0);

	/* set the subject of the new message */
	mw->subject = gtk_entry_new ();
	gtk_table_attach (GTK_TABLE(headers_page), mw->subject,
	                  1, 2, 3, 4,
			  GTK_FILL|GTK_EXPAND, GTK_FILL, GNOME_PAD, 0);

	gtk_signal_connect (GTK_OBJECT (mw->subject),
	                    "changed",
	                    update_title_with_subject,
	                    mw);
	/*-----
	 * Advanced Headers Page
	 * ---- */

	if (1)
	{
		gchar *pch;
		GtkWidget * table;
		GtkWidget * w = NULL;
		GtkWidget * eventbox = NULL; 
		int row = 0;

		table = adv_headers_page = gtk_table_new (4, 2, FALSE);

		w = gtk_label_new ( _("Followup-To:") );
		gtk_misc_set_alignment (GTK_MISC(w), 0.0, 0.5);
		eventbox = gtk_event_box_new ();
		gtk_container_add (GTK_CONTAINER(eventbox), w);
		gtk_tooltips_set_tip (GTK_TOOLTIPS(ttips), eventbox,
			_("The newsgroup or newsgroups where replies to "
			  "your posted message should go.  This is only "
			  "needed if it differs from the \"Post To Groups\" "
			  "header. "
			  "\nTo direct all replies to your email address, "
			  "use \"Followup-To: poster\""), "");
		gtk_table_attach (GTK_TABLE(table), eventbox, 0, 1, row, row+1,
				  GTK_FILL, GTK_FILL, GNOME_PAD, 0);
		mw->followup_to = gtk_entry_new ();
		gtk_table_attach (GTK_TABLE(table), mw->followup_to,
		                  1, 2, row, row+1,
				  GTK_FILL | GTK_EXPAND, GTK_FILL, GNOME_PAD, 0);

		++row;

		w = gtk_label_new ( _("Reply-To:") );
		gtk_misc_set_alignment (GTK_MISC(w), 0.0, 0.5);
		eventbox = gtk_event_box_new ();
		gtk_container_add (GTK_CONTAINER(eventbox), w);
		gtk_tooltips_set_tip (GTK_TOOLTIPS(ttips), eventbox,
			_("The e-mail account where mail replies to "
			  "your posted message should go.  This is only "
			  "needed if it differs from the \"From\" "
			  "header."), "");
		gtk_table_attach (GTK_TABLE(table), eventbox,
		                  0, 1, row, row+1,
				  GTK_FILL, GTK_FILL, GNOME_PAD, 0);
		mw->reply_to = gtk_entry_new ();
		pch = gnome_config_get_string ("/Pan/User/Reply_To");
		if (pch != NULL) {
			gtk_entry_set_text (GTK_ENTRY(mw->reply_to), pch);
			g_free (pch);
		}
		
		gtk_table_attach (GTK_TABLE(table), mw->reply_to,
		                  1, 2, row, row+1,
				  GTK_FILL | GTK_EXPAND, GTK_FILL, GNOME_PAD, 0);

		++row;

		w = gtk_label_new ( _("Organization:") );
		gtk_misc_set_alignment (GTK_MISC(w), 0.0, 0.5);
		eventbox = gtk_event_box_new ();
		gtk_container_add (GTK_CONTAINER(eventbox), w);
		gtk_tooltips_set_tip (GTK_TOOLTIPS(ttips), eventbox,
			_("The organization you're associated with."), "");
		gtk_table_attach (GTK_TABLE(table), eventbox,
		                  0, 1, row, row+1,
				  GTK_FILL, GTK_FILL, GNOME_PAD, 0);
		mw->organization = gtk_entry_new ();
		gtk_table_attach (GTK_TABLE(table), mw->organization,
		                  1, 2, row, row+1,
				  GTK_FILL|GTK_EXPAND, GTK_FILL, GNOME_PAD, 0);

		++row;
	}
	
	/*-----
	 * Fill the notebook
	 * ---- */
	
	notebook = gtk_notebook_new ();

        gtk_notebook_append_page (GTK_NOTEBOOK (notebook), headers_page,
			gtk_label_new (_("Headers")));
        gtk_notebook_append_page (GTK_NOTEBOOK (notebook), adv_headers_page,
			gtk_label_new (_("Advanced Headers")));
        gtk_notebook_append_page (GTK_NOTEBOOK (notebook), attachment_page(mw),
			gtk_label_new (_("Attachment")));

	pan_unlock();

	return notebook;
}

static void
populate_post_info_pane (MessageWindow *mw)
{
	if (mw->type == EDIT_ORIGINAL)
	{
		gchar * pch;
		const gchar * header;

		/* author */
		g_assert (mw->from!=NULL);
		pch = article_get_author_str (mw->article);
		if (is_nonempty_string(pch))
			gtk_entry_set_text (GTK_ENTRY(mw->from), pch);
		g_free (pch);

		/* subject */
		g_assert (mw->subject!=NULL);
		header = mw->article->subject;
		if (is_nonempty_string(header))
			gtk_entry_set_text (GTK_ENTRY(mw->subject), header);

		/* organization */
		g_assert (mw->organization!=NULL);
		header = article_get_header (mw->article, HEADER_ORGANIZATION);
		if (is_nonempty_string(header))
			gtk_entry_set_text (GTK_ENTRY(mw->organization), header);

		/* newsgroups */
		g_assert (mw->newsgroups!=NULL);
		header = article_get_header (mw->article, HEADER_NEWSGROUPS);
		if (is_nonempty_string(header))
			gtk_entry_set_text (GTK_ENTRY(mw->newsgroups), header);

		/* mail-to */
		g_assert (mw->users!=NULL);
		header = article_get_header (mw->article, PAN_MAIL_TO);
		if (is_nonempty_string(header))
			gtk_entry_set_text (GTK_ENTRY(mw->users), header);

		/* reply-to */
		g_assert (mw->reply_to!=NULL);
		header = article_get_header (mw->article, HEADER_REPLY_TO);
		if (is_nonempty_string(header))
			gtk_entry_set_text (GTK_ENTRY(mw->reply_to), header);

		/* followup-to */
		g_assert (mw->followup_to!=NULL);
		header = article_get_header (mw->article, HEADER_FOLLOWUP_TO);
		if (is_nonempty_string(header))
			gtk_entry_set_text (GTK_ENTRY(mw->followup_to), header);
	}
	else
	{
		Article * a = mw->article;
		Group *  group = NULL;
		gchar * p;

		/* figure out which group to use */
		if (a!=NULL) group = a->group;
		if (group==NULL) group = articlelist_get_group ();
		if (group==NULL) group = grouplist_get_selected_group ();

		/**
		***   FROM
		**/

		p = gnome_config_get_string ("/Pan/User/Email");
		if (p != NULL)
		{
			GString * use = g_string_new (NULL);
			gchar* real = gnome_config_get_string ("/Pan/User/Full_Name");

			g_strstrip (real);
			if (is_nonempty_string(real))
			{
				gboolean do_quote = *real!='"';

				if (do_quote)
					g_string_sprintfa (use, "\"%s\"", real);
				else
					g_string_sprintfa (use, "%s", real);

				g_string_append_c (use, ' ');
			}
			g_string_sprintfa (use, "<%s>", p);

			gtk_entry_set_text (GTK_ENTRY(mw->from), use->str);

			g_free (p);
			g_free (real);
			g_string_free (use, TRUE);
		}

		/**
		***  SUBJECT
		**/

		if (a!=NULL && is_nonempty_string(a->subject)) {
			gchar * subject = make_reply_string (a->subject);
			gtk_entry_set_text (GTK_ENTRY(mw->subject), subject);
			g_free (subject);
		}

		/**
		***  ORGANIZATION
		**/

		p = gnome_config_get_string ("/Pan/User/Organization");
		if (p != NULL) {
			gtk_entry_set_text (GTK_ENTRY(mw->organization), p);
			g_free (p);
		}

		/**
		***   NEWSGROUPS
		***   MAILTO
		**/

		p = NULL;

		/* are we posting an article? */
		if (a!=NULL && is_posting(mw->type))
		{
			const gchar * followup = article_get_header (a, HEADER_FOLLOWUP_TO);
			const gchar * newsgroups = article_get_header (a, HEADER_NEWSGROUPS);

			/* if user set followup-to, use that;
			 * otherwise, use newsgroups;
			 * otherwise, use the currently-selected group */
			if (is_nonempty_string(followup))
			{
				/* Followup-To: poster */
				if (!g_strcasecmp ("poster", followup)) /* explicit reply by mail */
				{
					const gchar * replyto = article_get_header (a, HEADER_REPLY_TO);

					gnome_ok_dialog_parented (
						_("``Followup-To: poster'': sending email to author."),
						GTK_WINDOW(mw->window));

					if (is_nonempty_string(replyto))
						gtk_entry_set_text (GTK_ENTRY(mw->users), replyto);
					else {
						gchar * pch = article_get_author_str (a);
						gtk_entry_set_text (GTK_ENTRY(mw->users), pch);
						g_free (pch);
					}
				}
				else /* explicit followup groups */
				{
					gtk_entry_set_text (GTK_ENTRY(mw->newsgroups), followup);
				}
			}
			else if (is_nonempty_string(newsgroups)) /* explicit groups */
			{
				gtk_entry_set_text (GTK_ENTRY(mw->newsgroups), newsgroups);
			}
			else /* implicit group */
			{
				if (group != NULL)
					gtk_entry_set_text (GTK_ENTRY(mw->newsgroups), group->name);
			}
		}
		else if (is_posting(mw->type))
		{
			if (group != NULL)
				gtk_entry_set_text (GTK_ENTRY(mw->newsgroups), group->name);
		}

		/* are we sending mail? */
		if (mw->type==EMAIL_REPLY || mw->type==EMAIL_AND_POST_REPLY)
		{
			/* figure out who to send mail to by checking reply-to: and from: */
			if (a!=NULL)
			{
				const gchar * replyto = article_get_header (a, HEADER_REPLY_TO);
				if (is_nonempty_string(replyto))
					gtk_entry_set_text (GTK_ENTRY(mw->users), replyto);
				else {
					gchar * pch = article_get_author_str (a);
					gtk_entry_set_text (GTK_ENTRY(mw->users), pch);
					g_free (pch);
				}
			}
		}
	}
}

static void
attachments_update_lines (MessageWindow *mw)
{
	struct stat statd;
   	char *filename = gtk_entry_get_text (GTK_ENTRY(mw->filename));

	if (filename && (0 == stat (filename, &statd))) {
		char *p;
		unsigned long lines;
		unsigned long lines_per_part;

		lines = statd.st_size / 45;
		p = g_strdup_printf ("%lu", lines);
		gtk_label_set_text (GTK_LABEL(mw->lines), p); 
		g_free (p);
		lines_per_part = gtk_spin_button_get_value_as_int (
	                      GTK_SPIN_BUTTON(mw->lines_per_part));
		p = g_strdup_printf ("%lu", (lines / lines_per_part) + 1);
		gtk_label_set_text (GTK_LABEL(mw->parts), p); 
		g_free (p);
	}
}

static void
attachments_lines_per_part_changed (GtkWidget *w, MessageWindow *mw)
{
	attachments_update_lines (mw);
}

static void
attachments_page_file_selected (GtkWidget *w, GtkWidget *f)
{
	char *selected_filename = gtk_file_selection_get_filename (GTK_FILE_SELECTION(f));
	MessageWindow * mw = gtk_object_get_user_data (GTK_OBJECT(f));
   	gtk_entry_set_text (GTK_ENTRY(mw->filename), selected_filename);
	attachments_update_lines (mw);
}

static void
attachment_file_selector (GtkWidget *menuitem, MessageWindow * mw) 
{
	GtkWidget *file_selector;

	/* Create the selector */
	file_selector = gtk_file_selection_new(_("Select the file to attach."));
   	gtk_object_set_user_data (GTK_OBJECT(file_selector), mw);

	gtk_signal_connect (
		GTK_OBJECT (GTK_FILE_SELECTION(file_selector)->ok_button),
		"clicked", GTK_SIGNAL_FUNC (attachments_page_file_selected), 
		(gpointer) file_selector);
                            
	/* Ensure that the dialog box is destroyed
	   when the user clicks a button. */
    	gtk_signal_connect_object (
		GTK_OBJECT(GTK_FILE_SELECTION(file_selector)->ok_button),
		"clicked", GTK_SIGNAL_FUNC (gtk_widget_destroy),
		(gpointer) file_selector);
	gtk_signal_connect_object (GTK_OBJECT (
		GTK_FILE_SELECTION(file_selector)->cancel_button),
		"clicked", GTK_SIGNAL_FUNC (gtk_widget_destroy),
		(gpointer) file_selector);
    
	/* Display that dialog */
    	gtk_widget_show (file_selector);
}

static void
append_sig_file (GtkWidget * text)
{
	gchar * sigfile = gnome_config_get_string ("/Pan/User/Signature_File");
	if (is_nonempty_string (sigfile))
	{
		gchar * tmp = pan_substitute (sigfile,"~",g_get_home_dir());
		GArray * sig = read_file (tmp);
		if (sig!=NULL)
		{
			/* insert the "-- " if it's not already there */
			if (strstr(sig->data,  "\n-- ") == NULL)
				gtk_text_insert (GTK_TEXT(text),
					NULL, &text_fg_color,
					NULL, "\n-- \n\n", 5);

			/* insert the signature */
			gtk_text_insert (GTK_TEXT(text),
				NULL, &text_fg_color,
				NULL, sig->data, sig->len);
			gtk_text_set_point (GTK_TEXT(text), 0);
			g_array_free (sig, TRUE);
		}
		g_free (tmp);
	}
	g_free (sigfile);
}


static GtkWidget*
create_post_body_pane (MessageWindow * mw)
{
	GtkWidget * scrolled_window;
	GtkStyle * style;

	g_return_val_if_fail (mw!=NULL, NULL);

	pan_lock();

	mw->body = gtk_text_new (NULL, NULL);

	/* set the text widget's properties */
	widget_set_font (GTK_WIDGET(mw->body), message_body_font);
	style = gtk_widget_get_style(mw->body);
	style->text[0] = text_fg_color;
	style->base[0] = text_bg_color;
	gtk_text_set_word_wrap (GTK_TEXT (mw->body), FALSE);
	gtk_text_set_line_wrap (GTK_TEXT (mw->body), FALSE);
	gtk_text_set_editable (GTK_TEXT (mw->body), TRUE);

	gtk_text_freeze (GTK_TEXT (mw->body));

	/* insert message body */
	if (mw->type==NNTP_REPLY
		|| mw->type==EMAIL_REPLY
		|| mw->type==EMAIL_AND_POST_REPLY
		|| mw->type==EMAIL_FORWARD)
	{
		gchar * body;
		gchar * attribution;
		const gchar * march; 
		GString * line = g_string_new (NULL);

		/* precede quoted material with attribution
		   line containing original author */
		pan_warn_if_fail (mw->article!=NULL);
		attribution = article_get_attribution_string (mw->article);
		if (is_nonempty_string(attribution))
		{
			g_string_sprintf (line, "%s\n\n", attribution);
			gtk_text_insert (GTK_TEXT (mw->body),
					 NULL, &text_fg_color, NULL,
					 line->str, -1);
		}
		replace_gstr (&attribution, NULL);

		/* get the entire original message, or the portion
		   of it that we're replying to */
		body = g_strdup (article_get_header (mw->article, "REPLY_PORTION"));
		if (body == NULL)
			body = article_get_body (mw->article);

		/* quote the original message.
		   It's good practice for newsreaders to use '> ' as the quote
		   prefix for newly-quoted, and '>' for repetitively quoted
		   text.   Also, included text should NOT include the original
		   article's signature unless explicitly requested by the user.
		   Finally, strip out the carriage returns as the only thing
		   they do is show up badly on some fonts. */
	       	g_string_truncate (line, 0);
		replace_gstr (&body, pan_substitute(body, "\r", ""));
		march = body;
		while (get_next_token_g_str(march, '\n', &march, line))
		{
			const gchar * prefix = *line->str=='>' ? ">" : "> ";
			g_string_prepend (line, prefix);
			g_string_append_c (line, '\n');
			gtk_text_insert (GTK_TEXT (mw->body), NULL,
			                 &text_quoted_color, NULL,
			                 line->str, line->len);
		}

		gtk_text_insert (GTK_TEXT (mw->body), NULL,
		                 &text_fg_color, NULL,
		                 "\n\n", 1);


		/*FIXME: this works but is kind of hackish */
		if (fill_body_on_reply)
		{
			char* body = gtk_editable_get_chars (GTK_EDITABLE(mw->body), 0, -1);
			char* new_body = fill_body (body);

			pan_unlock();
			update_body_pane (mw->body, new_body);
			pan_lock();

			g_free (body);
			g_free (new_body);
		}

		g_free (body);
		g_string_free (line, TRUE);
	}
	
	/* Add something the color gets set back for the user's input */
	gtk_text_insert (GTK_TEXT (mw->body), NULL, &text_fg_color, NULL, "\n", 1);
	append_sig_file (mw->body);

	gtk_text_thaw (GTK_TEXT (mw->body));

	scrolled_window = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_window),
					GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_container_add (GTK_CONTAINER (scrolled_window), mw->body);

	pan_unlock();

	return scrolled_window;
}


/*---[ create_read_body_pane ]----------------------------------------
 * will be called from Reading code
 * This should NOT be called within a pan lock
 *--------------------------------------------------------------------*/
static GtkWidget *
create_read_body_pane (GtkWidget *body_widget,
                       MessageWindowType type,
                       Article * article)
{
	pan_lock();

	/* instantiate the body widget if we need to */
	if (!body_widget)
	{
		GtkStyle * style;

		body_widget = gtk_text_new (NULL, NULL);

		/*gtk_text_set_word_wrap (GTK_TEXT (body_widget), TRUE);*/
		widget_set_font (body_widget, message_body_font);
		style = gtk_widget_get_style(body_widget);
		style->text[0] = text_fg_color;
		style->base[0] = text_bg_color;
	}

	gtk_text_freeze (GTK_TEXT (body_widget));

	/* insert message body */
	if (type==NNTP_READ || type==EDIT_ORIGINAL)
	{
		int i;
		gchar * body;
		gchar **sd;

		body = article_get_body (article);
		sd = is_nonempty_string(body) ? g_strsplit (body, "\n", -1) : NULL;

		for (i=0; sd!=NULL && sd[i]; ++i)
		{
			gchar *line = sd[i];
			size_t len = strlen(line);
			GdkColor *pFG = *line=='>'
				? &text_quoted_color
				: &text_fg_color;
			if (line[len-1] == '\r') /* omit carriage returns */
				line[--len] = '\0';
			line[len++] = '\n';
			gtk_text_insert (GTK_TEXT(body_widget),
			                 NULL, pFG, NULL, line, len);
		}

		gtk_text_set_point (GTK_TEXT(body_widget), 0);

		/* cleanup */
		g_free (body);
		g_strfreev (sd);
	}

	gtk_text_thaw (GTK_TEXT (body_widget));
	gtk_widget_show (body_widget);

	pan_unlock ();

	return body_widget;
}


/*---[ update_body_pane ]---------------------------------------------
 * will only be called for messages being "read"
 *--------------------------------------------------------------------*/
void
update_body_pane (GtkWidget    * body_widget,
                  const char   * body)
{
	GtkText *text = NULL;

	/* sanity checks */
	g_return_if_fail (body_widget!=NULL);
	g_return_if_fail (GTK_IS_TEXT(body_widget));

	text = GTK_TEXT(body_widget);

	pan_lock();
	gtk_text_set_point (text, 0);
	gtk_editable_delete_text (GTK_EDITABLE(body_widget), 0, -1);
	if (is_nonempty_string(body))
	{
		/* insert the article body */
	       	GString * line = g_string_new (NULL);
		gtk_text_freeze (text);
		while (get_next_token_g_str(body, '\n', &body, line))
		{
			GdkColor * pFG;

			/* what color to use? */
			pFG = line->len && line->str[0]=='>'
				? &text_quoted_color
				: &text_fg_color;

			/* strip carriage return */
			if (line->len!=0 && line->str[line->len-1]=='\r')
				g_string_erase (line, line->len-1, 1);

			/* add linefeed */
			g_string_append_c (line, '\n');

			/* insert this line */
			gtk_text_insert (text, NULL, pFG, NULL, line->str, line->len);
		}
		gtk_text_thaw (text);
		g_string_free (line, TRUE);
	}
	pan_unlock();
}

void
message_post_window (GtkWidget *widget, Article * article)
{
	message_window_new (article, NNTP_POST);
}


static void
message_post_window_create (MessageWindow *mw)
{
	GtkWidget *vbox = NULL;
	GtkWidget *info_post_pane = NULL;
	GtkWidget *body_post_pane = NULL;

	pan_lock();
	mw->window = gnome_app_new ("Post", "New Message");
	gtk_signal_connect (GTK_OBJECT(mw->window), "destroy",
	                    GTK_SIGNAL_FUNC(window_destroy_cb), mw);
	vbox = gtk_vbox_new (FALSE, 0);
	gnome_app_set_contents (GNOME_APP (mw->window), vbox);
	gtk_window_set_default_size (GTK_WINDOW (mw->window), 550, 410);
	gnome_app_create_menus_with_data (GNOME_APP (mw->window),
	                                  message_post_main_menu,
	                                  mw);
	gnome_app_create_toolbar_with_data (GNOME_APP (mw->window),
	                                    message_post_toolbar,
	                                    mw);
	pan_unlock();

	info_post_pane = create_post_info_pane (mw);
	populate_post_info_pane (mw);
	body_post_pane = create_post_body_pane(mw);

	pan_lock();
	gtk_box_pack_start (GTK_BOX(vbox), info_post_pane, FALSE, FALSE, GNOME_PAD_SMALL);
	gtk_box_pack_start (GTK_BOX(vbox), body_post_pane, TRUE, TRUE, 0);
	pan_unlock();
}


void
message_read_window (GtkWidget *widget, Article * article)
{
	if (article == NULL)
		article = articlelist_get_selected_article ();

	if (article != NULL)
		message_window_new (article, NNTP_READ);
}


/*--------------------------------------------------------------------
 *
 *--------------------------------------------------------------------*/
static void
message_read_window_create (MessageWindow *mw)
{
	GtkWidget *vbox = NULL;
	GtkWidget *read_body_pane = NULL;
	GtkWidget *scrolled_window = NULL;

	/* create a new window */
	pan_lock();
	mw->window = gnome_app_new ("Read", mw->article->subject);
	vbox = gtk_vbox_new (FALSE, 0);
	gnome_app_set_contents (GNOME_APP (mw->window), vbox);
	gtk_window_set_default_size (GTK_WINDOW (mw->window), 550, 410);
	gnome_app_create_menus_with_data (GNOME_APP(mw->window),
	                                  message_read_main_menu, mw);
	gnome_app_create_toolbar_with_data (GNOME_APP(mw->window),
	                                    message_read_toolbar, mw);
	gtk_signal_connect (GTK_OBJECT(mw->window), "destroy",
	                    GTK_SIGNAL_FUNC(window_destroy_cb), mw);
	pan_unlock();

	/* create read info pane */
	mw->read_info_pane = create_read_info_pane (mw);

	/* create read body pane */
	read_body_pane = create_read_body_pane (NULL, mw->type, mw->article);
	pan_lock();

	/* add info pane */
	gtk_box_pack_start (
		GTK_BOX (vbox), mw->read_info_pane,
		FALSE, FALSE, GNOME_PAD_SMALL);
	gtk_box_pack_start (
		GTK_BOX (vbox), gtk_hseparator_new(),
		FALSE, FALSE, 2);

	/* add scrolled body window */
	scrolled_window = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_policy (
		GTK_SCROLLED_WINDOW (scrolled_window),
		GTK_POLICY_AUTOMATIC,
		GTK_POLICY_AUTOMATIC);
	gtk_container_add (
		GTK_CONTAINER (scrolled_window),
		read_body_pane);
	gtk_box_pack_start (
		GTK_BOX(vbox), scrolled_window,
		TRUE, TRUE, 0);
		
	pan_unlock();
}


/*--------------------------------------------------------------------
 *
 *--------------------------------------------------------------------*/

void
message_reply_external (GtkWidget    * widget,
                        Article      * article)
{
	gchar * sub = NULL;
	gchar * body = NULL;
	gchar * author = NULL;

	/* What are we replying to anyway ? */
	if (article == NULL)
		article = articlelist_get_selected_article ();
	if (article != NULL) {
		sub = make_reply_string (article->subject);
		author = article_get_author_str (article);
		body = acache_load_body (article->message_id);
	}

	/* if we are replying to something, then go ahead and reply... */
	if (body!=NULL && author!=NULL && sub!=NULL)
	{
		GString * exec = g_string_new (external_mailer);

		/* If %t is being used, write the body to a temporary file */
		if ((strstr(external_mailer, "%t")) != NULL)
		{
			FILE * f = NULL;
			gchar * filename = NULL;
			filename = pan_make_temp (&f);
			if (filename != NULL)
			{
				/* Quote the body */
				gint i;
				gchar ** sd = g_strsplit (body, "\n", -1);
				GString * tmp = g_string_new (NULL);

				for (i=0; sd[i]; ++i)
				{
					const gchar * prefix =
						(*sd[i]=='>') ? ">" : "> ";
					size_t len = strlen(sd[i]);

					/* trim out linefeeds */
					if (len && (sd[i][len-1]=='\r'))
						sd[i][--len] = '\0';

					g_string_sprintf (
						tmp, "%s%s\n", prefix, sd[i]);
					fputs (tmp->str, f);
				}

				pan_g_string_replace (exec, "%t", filename);

				/* cleanup */
				g_free (filename);
				g_string_free (tmp, TRUE);
				g_strfreev (sd);
				fclose (f);
			}
		}

		/* Expand the rest of the command string */
		pan_g_string_replace (exec, "%r", author);
		pan_g_string_replace (exec, "%s", sub);

		/* Launch the command */
		gnome_execute_shell (NULL, exec->str);

		g_string_free (exec, TRUE);
	}
	
	/* clean up */
	g_free (sub);
	g_free (author);
	g_free (body);
}

static void
make_reply_window (Article * article, MessageWindowType type)
{
	gchar * body;
	gulong end_pos;
	gulong start_pos;
	GtkEditable * editable;

	g_return_if_fail (article!=NULL);
	g_return_if_fail (type==EMAIL_REPLY
		|| type==NNTP_REPLY
		|| type==EMAIL_AND_POST_REPLY
		|| type==EMAIL_FORWARD);

	/* if this isn't the same article in the article reading window,
	 * then fire up a task to retreive the body from the local cache
	 * or on the net.  FIXME task-body really needs to be decoupled
	 * from viewing.
	 */
	if (article != current_article) {
		text_set_from_article (article, type, TRUE);
		return;
	}

	/* At this point we know we're replying to the article in the
	   main message window.  We check to see if the user's got any
	   of that message selected; if it is, we will only include that
	   portion of the message in the follow-up. (gnksa 10e).
	   Also, strip out the signature file here. (gnksa 10d). */
	editable = GTK_EDITABLE(Pan.text);
	start_pos = editable->has_selection ? editable->selection_start_pos : 0;
	end_pos = editable->has_selection ? editable->selection_end_pos : -1;
	body = gtk_editable_get_chars (editable, start_pos, end_pos);
	replace_gstr (&body, pan_substitute(body, "\r", ""));
	if (!editable->has_selection) {
		gchar * pch = strstr (body, "-- \n");
		if (pch != NULL)
			*pch = '\0';
	}
	article_set_header_pair_chunk (
		article, "REPLY_PORTION",
		group_chunk_string(article->group,body));

	/* open up a reply window with the specified body. */
	message_window_new (article, type);
	g_free (body);
}

void
message_forward_window (GtkWidget * widget)
{
	Article * a = articlelist_get_selected_article ();
	if (a != NULL)
		make_reply_window (a, EMAIL_FORWARD);
}

void
message_edit_window (GtkWidget * widget, Article * a)
{
	if (a == NULL)
		a = articlelist_get_selected_article ();
	if (a != NULL)
		message_window_new (a, EDIT_ORIGINAL);
}

void
message_reply_window (GtkWidget * widget)
{
	Article * a = articlelist_get_selected_article ();
	if (a != NULL)
		make_reply_window (a, EMAIL_REPLY);
}

void
message_followup_reply_window (GtkWidget * widget)
{
	Article * a = articlelist_get_selected_article ();
	if (a != NULL)
		make_reply_window (a, EMAIL_AND_POST_REPLY);
}


void
message_followup_window (GtkWidget * widget)
{
	Article * a = articlelist_get_selected_article ();
	if (a != NULL)
		make_reply_window (a, NNTP_REPLY);
}

static void
message_edit_window_create (MessageWindow *mw)
{
	GtkWidget * v;
	GtkWidget * post_info_pane;
	GtkWidget * post_body_pane;

	/* create the window */
	pan_lock ();
	mw->window = gnome_app_new ("Edit", mw->article->subject);
	v = gtk_vbox_new (FALSE, 0);
	gnome_app_set_contents (GNOME_APP(mw->window), v);
	gtk_window_set_default_size (GTK_WINDOW (mw->window), 550, 410);
	gtk_signal_connect (GTK_OBJECT(mw->window), "destroy",
	                    GTK_SIGNAL_FUNC(window_destroy_cb), mw);
	pan_unlock ();

	post_info_pane = create_post_info_pane (mw);
	populate_post_info_pane (mw);
	post_body_pane = create_read_body_pane (NULL, mw->type, mw->article);

	pan_lock ();
	gtk_text_set_editable (GTK_TEXT(mw->body=post_body_pane), TRUE);
	gnome_app_create_menus_with_data (GNOME_APP(mw->window),
	                                  message_edit_main_menu,
	                                  mw);
	gnome_app_create_toolbar_with_data (GNOME_APP(mw->window),
	                                    message_edit_toolbar,
	                                    mw);
	gtk_box_pack_start (GTK_BOX(v), post_info_pane, 0, 0, GNOME_PAD_SMALL);
	gtk_box_pack_start (GTK_BOX(v), gtk_hseparator_new(), 0, 0, 2);
	gtk_box_pack_start (GTK_BOX(v), post_body_pane, 1, 1, 0);
	pan_unlock();
}

static void
message_reply_window_create (MessageWindow *mw)
{
	GtkWidget * v;
	GtkWidget * post_info_pane;
	GtkWidget * post_body_pane;
	gchar * title = make_reply_string (mw->article->subject);

	/* create the window */
	pan_lock();
	mw->window = gnome_app_new ("Reply", title);
	v = gtk_vbox_new (FALSE, 0);
	gnome_app_set_contents (GNOME_APP (mw->window), v);
	gtk_window_set_default_size (GTK_WINDOW (mw->window), 550, 410);
	gtk_signal_connect (GTK_OBJECT(mw->window), "destroy",
	                    GTK_SIGNAL_FUNC(window_destroy_cb), mw);
	pan_unlock();

	post_info_pane = create_post_info_pane (mw);
	populate_post_info_pane (mw);
	post_body_pane = create_post_body_pane (mw);

	pan_lock ();
	gnome_app_create_menus_with_data (GNOME_APP(mw->window),
	                                  message_post_main_menu,
	                                  mw);
	gnome_app_create_toolbar_with_data (GNOME_APP(mw->window),
	                                    message_post_toolbar,
	                                    mw);
	gtk_box_pack_start (GTK_BOX(v), post_info_pane, 0, 0, GNOME_PAD_SMALL);
	gtk_box_pack_start (GTK_BOX(v), gtk_hseparator_new(), 0, 0, 2);
	gtk_box_pack_start (GTK_BOX(v), post_body_pane, 1, 1, 0);
	pan_unlock();

	/* cleanup */
	g_free (title);
}

void
message_window_new (Article * article, MessageWindowType type)
{
	MessageWindow *mw;

	if (article != NULL)
		group_ref_articles (article->group, NULL);

	mw = g_new0 (MessageWindow, 1);
	mw->article = article;
	mw->type = type;

	switch (type)
	{
		case NNTP_READ:
			message_read_window_create (mw);
			break;

		case NNTP_POST:
			message_post_window_create (mw);
			break;

		case NNTP_REPLY:
		case EMAIL_REPLY:
		case EMAIL_AND_POST_REPLY:
		case EMAIL_FORWARD:
			message_reply_window_create (mw);
			break;

		case EDIT_ORIGINAL:
			message_edit_window_create (mw);
			break;

		default:
			break;
	}

	if (mw->window != NULL)
		gui_popup_draw (mw->window, Pan.window);
}

/***
****
****  SENDING THE MESSAGE
****
***/

static gchar*
get_text_from_editable (GtkWidget* w)
{
	gchar * pch;
	g_return_val_if_fail (GTK_IS_EDITABLE(w), NULL);
	pch = gtk_editable_get_chars (GTK_EDITABLE(w), 0, -1);
	g_strstrip (pch);
	return pch;
}

static gchar*
get_text_from_maybe_editable (GtkWidget * w)
{
	gchar * pch = NULL;

	if (w!=NULL && GTK_IS_EDITABLE(w))
	{
		pch = gtk_editable_get_chars (GTK_EDITABLE(w), 0, -1);
		g_strstrip (pch);
	}

	return pch;
}

static gchar*
message_window_get_lines_per_part (const MessageWindow* mw)
{
	int lpp; 

	g_return_val_if_fail (mw!=NULL, NULL);
	
	lpp = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON(mw->lines_per_part));

	return g_strdup_printf ("%d", lpp);
}


static gchar*
message_window_get_filename (const MessageWindow* mw)
{
	gchar * pch;

	g_return_val_if_fail (mw!=NULL, NULL);

	pch = get_text_from_maybe_editable (mw->filename);

	return pch;
}

static gchar*
message_window_get_reply_to (const MessageWindow* mw)
{
	gchar * pch;

	g_return_val_if_fail (mw!=NULL, NULL);

	pch = get_text_from_maybe_editable (mw->reply_to);
	if (pch == NULL)
		pch = g_strdup ("");

	return pch;
}

static gchar*
message_window_get_group_str (GtkWidget * e)
{
	gchar * pch;

	g_return_val_if_fail (e!=NULL, NULL);

	pch = get_text_from_maybe_editable (e);
	if (pch == NULL)
	{
		pch = g_strdup ("");
	}
	else
	{
		gchar * tmp;

		/* ensure all the delimiters are commas */
		g_strdelimit (pch, ":; ", ',');

		/* remove leading/trailing commas */
		for (tmp=pch; *tmp==','; ++tmp);
		g_memmove (pch, tmp, strlen(tmp)+1);
		for (tmp=pch+strlen(pch)-1; tmp>pch && *tmp==','; --tmp);
		tmp[1] = '\0';

		/* remove empty entries */
		while ((tmp=strstr(pch,",,")) != NULL)
			g_memmove (tmp, tmp+1, strlen(tmp+1)+1);
	}

	return pch;
}

static gchar*
message_window_get_organization (const MessageWindow* mw)
{
	gchar * pch = NULL;
	g_return_val_if_fail (mw!=NULL, NULL);
	pch = get_text_from_maybe_editable (mw->organization);
	if (pch == NULL)
		pch = gnome_config_get_string ("/Pan/User/Organization");
	if (pch == NULL)
		pch = g_strdup ("");
	if (pch != NULL)
		g_strstrip (pch);
	return pch;
}

static gchar*
message_window_get_reverse_path (const MessageWindow* mw)
{
	gchar * from = get_text_from_editable (mw->from);
	gchar * address = NULL;
	gchar * realname = NULL;
	gchar * retval = NULL;
	int status = gnksa_do_check_from (from, &address, &realname);
	g_free (from);
	g_free (realname);
	if (status == GNKSA_OK)
		retval = address;
	else
		g_free (address);
	return retval;
}

static void
article_add_attachment (Article *a, MessageWindow *mw)
{
	gchar *pch;

	if ((pch = message_window_get_filename(mw)) != NULL) {

		/* filename */
		article_set_header_pair_chunk (a, PAN_ATTACH_FILE, pch); 
		/* lines_per_part */
		pch = message_window_get_lines_per_part(mw);
		article_set_header_pair_chunk (a, PAN_LINES_PER_PART, pch); 
	}
}

static void
populate_article_from_mw (Article * article, MessageWindow * mw)
{
	gchar * pch;

	/* date */
	article->date = time(0);

	/* group's article number */
	article->number = 1;

	/* author */
	pch = get_text_from_editable (mw->from);
	article_set_author_from_header (article, pch);
	g_free (pch);

	/* subject */
	pch = get_text_from_editable (mw->subject);
	article_set_string_field (article, &article->subject, pch);
	g_free (pch);

	/* references (for posting articles) */
	if (mw->article != NULL) {
		pch = gnksa_generate_references (mw->article->references,
						 mw->article->message_id);
		if (pch != NULL) {
			article_set_string_field (article, &article->references, pch);
			g_free (pch);
		}
	}

	/* organization */
	pch = message_window_get_organization(mw);
	article_set_header_pair_chunk (article,
		HEADER_ORGANIZATION,
		group_chunk_string(article->group,pch));
	g_free (pch);

	/* followup-to */
	pch = message_window_get_group_str(mw->followup_to);
	article_set_header_pair_chunk (article,
		HEADER_FOLLOWUP_TO,
		group_chunk_string(article->group,pch));
	g_free (pch);

	/* reply-to */
	pch = message_window_get_reply_to (mw);
	article_set_header_pair_chunk (article,
		HEADER_REPLY_TO,
		group_chunk_string(article->group,pch));
	g_free (pch);

	/* pan-body */
	pch = get_text_from_editable (mw->body);
	article_set_header_pair_chunk (article,
		PAN_BODY,
		group_chunk_string(article->group,pch));
	g_free (pch);

	/* newsgroups (for posting articles) */
	pch = message_window_get_group_str (mw->newsgroups);
	article_set_header_pair_chunk (article,
		HEADER_NEWSGROUPS,
		group_chunk_string(article->group,pch));
	g_free (pch);

	/* pan-mail-to (for sending mail) */
	pch = get_text_from_editable (mw->users);
	article_set_header_pair_chunk (article,
		PAN_MAIL_TO,
		group_chunk_string(article->group,pch));
	g_free (pch);

	/* pan-reverse-path (for sending mail) */
	pch = message_window_get_reverse_path (mw);
	if (pch != NULL) {
		article_set_header_pair_chunk (article,
			PAN_REVERSE_PATH,
			group_chunk_string(article->group,pch));
		g_free (pch);
	}

	/* attribution header (for message-check) */
	if (mw->article!=NULL) {
		gchar * attribution = article_get_attribution_string (mw->article);
		if (is_nonempty_string(attribution))
			article_set_header_pair_chunk (article,
				"ATTRIBUTION",
				group_chunk_string(article->group,attribution));
		g_free (attribution);
	}
}

static Article*
build_article_from_window (MessageWindow * mw)
{
	Group * folder = folder_get_by_name (PAN_SENDLATER);
	Article * article = article_new (folder);
	populate_article_from_mw (article, mw);
	return article;
}

static void
save_cb (GtkWidget * button, MessageWindow * mw)
{
	Article * a = mw->article;
	Article * tmp;
	const gchar * pch_old;
	const gchar * pch_new;

	/* update the article fields from changes made in the gui */
	tmp = article_new (a->group);
	populate_article_from_mw (tmp, mw);
	a->author_addr = tmp->author_addr;
	a->author_real = tmp->author_real;
	a->subject = tmp->subject;

#define update_header(key) \
	pch_old = article_get_header (tmp, HEADER_ORGANIZATION); \
	pch_new = article_get_header (  a, HEADER_ORGANIZATION); \
	if (!is_nonempty_string(pch_new)) \
		article_remove_header (a, HEADER_ORGANIZATION); \
	else if (pan_strcmp (pch_old, pch_new)) \
		article_set_header_pair_chunk (a, HEADER_ORGANIZATION, pch_new);

	update_header (HEADER_ORGANIZATION)
	update_header (HEADER_FOLLOWUP_TO)
	update_header (HEADER_NEWSGROUPS)
	update_header (HEADER_REPLY_TO)
	update_header (PAN_BODY)
	update_header (PAN_MAIL_TO)
	update_header (PAN_REVERSE_PATH)
	update_header ("ATTRIBUTION")

#undef update_header

	pan_object_unref (PAN_OBJECT(tmp));

	message_window_destroy (mw);
}

static void
send_now_cb (GtkWidget * button, MessageWindow * mw)
{
	Article * article = build_article_from_window (mw);
	Server * server = grouplist_get_server ();
	
	article_add_attachment (article, mw);

	if (!queue_article_for_posting (server, article, TRUE))
		message_window_destroy (mw);
}

static void
window_destroy_cb (GtkWidget * window, MessageWindow * mw)
{
	g_message ("message_window_destroy");
	if (mw->article != NULL) {
		g_message ("unreffing");
		group_unref_articles (mw->article->group, NULL);
	}
	g_free (mw);
}

static void
send_later_cb (GtkWidget * button, MessageWindow * mw)
{
	Article * article = build_article_from_window (mw);
	Server * server = grouplist_get_server ();
	GPtrArray * errors = g_ptr_array_new ();
	GoodnessLevel goodness = OKAY;

	article_add_attachment (article, mw);

	/* if the article's okay, then store it for later sending */
	check_article_and_prompt (article, server, errors, &goodness);
	if (goodness == OKAY)
	{
		Group * folder;

		/* set the new article's relationship with sendlater */
		article_set_header_pair_chunk (article,
			PAN_SERVER,
			group_chunk_string(article->group, server->name));
		folder = folder_get_by_name (PAN_SENDLATER);
		group_add_article (folder, article);

		/* close this window */
		message_window_destroy (mw);
	}

	/* cleanup */
	pan_g_ptr_array_foreach (errors, (GFunc)g_free, NULL);
	g_ptr_array_free (errors, TRUE);
}

/**
***  Save as File
**/

static void
message_window_save_ok_clicked (GtkWidget * widget, GtkWidget * file_entry)
{
	const gchar * filename;
	gchar * writeme = NULL;
	const MessageWindow * mw;

	/* get information from the widget */
	pan_lock();
	mw = (const MessageWindow*) gtk_object_get_data (GTK_OBJECT(file_entry), "mw");
	filename = gtk_file_selection_get_filename (GTK_FILE_SELECTION (file_entry));
	pan_unlock();

	/* based on whether it's a reply or not, use acache or current */
	if (mw->type == NNTP_READ)
	{
		const gchar * msg_id = mw->article->message_id;
		gchar * body = acache_load_body (msg_id);
		gchar * header = acache_load_header (msg_id);
		writeme = g_strconcat (header, "\n\n", body);
		g_free (header);
		g_free (body);
	}
	else
	{
		writeme = get_text_from_editable (mw->body);
	}

	/* write the text */
	replace_gstr (&writeme, pan_substitute(writeme,"\r",""));
	if (is_nonempty_string(writeme))
	{
		FILE * msg_dump = fopen (filename, "w");
		if (msg_dump != NULL)
		{
			fprintf (msg_dump, "%s\n", writeme);
			fclose (msg_dump);

			pan_lock();
			gtk_widget_destroy (file_entry);
			pan_unlock();
		}
		else
		{
			g_message ("fopen() failed to save message");
		}
	}

	/* cleanup */
	g_free (writeme);
}

static void
message_window_save_cb (GtkWidget *widget, MessageWindow * mw)
{
	GtkWidget *file_entry = NULL;
	
	pan_lock();
	
	file_entry = gtk_file_selection_new (_("Save message to file"));

	gtk_file_selection_hide_fileop_buttons (GTK_FILE_SELECTION (file_entry));

	gtk_signal_connect_object (GTK_OBJECT (GTK_FILE_SELECTION (file_entry)->cancel_button),
				   "clicked",
				   GTK_SIGNAL_FUNC (gtk_widget_destroy),
				   GTK_OBJECT (file_entry));

	/* message_id is const, but we don't alter it here */
	gtk_object_set_data (GTK_OBJECT (file_entry), "mw", (gpointer)mw);

	gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (file_entry)->ok_button),
				"clicked",
				GTK_SIGNAL_FUNC (message_window_save_ok_clicked),
				file_entry);

	gtk_widget_show_all (file_entry);
	pan_unlock();
}
